@@ -52,9 +52,18 @@ sub init {
   $this->{CLASS}     = $args{class};
   $this->{CATCH}     = $args{catch};
   $this->{CONDITION} = $args{condition};
+  $this->{ALIAS}     = $args{alias};
+  $this->{TAGS}      = $args{tags};
+  $this->{EMIT_CONDITION} = $args{emit_condition};
 
-  if (ref($this->{CATCH})
-      and @{$this->{CATCH}} > 1
+  my $index = 0;
+  foreach my $arg ( @{$this->{ARGUMENTS}} ) {
+      $arg->{FUNCTION} = $this;
+      $arg->{INDEX} = $index;
+      ++$index;
+  }
+
+  if (@{$this->catch} > 1
       and grep {$_ eq 'nothing'} @{$this->{CATCH}})
   {
     Carp::croak( ref($this) . " '" . $this->{CPP_NAME}
@@ -75,14 +84,15 @@ and stores a reference to those objects.
 
 sub resolve_typemaps {
   my $this = shift;
+  my $index = 0;
 
   if( $this->ret_type ) {
-    $this->{TYPEMAPS}{RET_TYPE} =
+    $this->{TYPEMAPS}{RET_TYPE} ||=
       ExtUtils::XSpp::Typemap::get_typemap_for_type( $this->ret_type );
   }
   foreach my $a ( @{$this->arguments} ) {
-    my $t = ExtUtils::XSpp::Typemap::get_typemap_for_type( $a->type );
-    push @{$this->{TYPEMAPS}{ARGUMENTS}}, $t;
+    $this->{TYPEMAPS}{ARGUMENTS}[$index++] ||=
+      ExtUtils::XSpp::Typemap::get_typemap_for_type( $a->type );
   }
 }
 
@@ -97,7 +107,7 @@ the C<%catch> directives associated with this function.
 sub resolve_exceptions {
   my $this = shift;
 
-  my @catch = @{$this->{CATCH} || []};
+  my @catch = @{$this->catch};
 
   my @exceptions;
 
@@ -142,13 +152,12 @@ sub add_exception_handlers {
   my $this = shift;
 
   # ignore class %catch'es if overridden with "nothing" in the method
-  if ($this->{CATCH} and @{$this->{CATCH}} == 1
-      and $this->{CATCH} eq 'nothing') {
+  if (@{$this->catch} == 1 and $this->{CATCH}[0] eq 'nothing') {
     return();
   }
 
   # ignore class %catch{nothing} if overridden in the method
-  if (@_ == 1 and $_[0] eq 'nothing' and @{$this->{CATCH}}) {
+  if (@_ == 1 and $_[0] eq 'nothing' and @{$this->catch}) {
     return();
   }
 
@@ -158,23 +167,6 @@ sub add_exception_handlers {
   return();
 }
 
-
-=head2 argument_style
-
-Returns either C<ansi> or C<kr>. C<kr> is the default.
-C<ansi> is returned if any one of the arguments uses the XS
-C<length> feature.
-
-=cut
-
-sub argument_style {
-  my $this = shift;
-  foreach my $arg (@{$this->{ARGUMENTS}}) {
-    return 'ansi' if $arg->name =~ /length.*\(/;
-  }
-  return 'kr';
-}
-
 # Depending on argument style, this produces either: (style=kr)
 #
 # return_type
@@ -183,6 +175,8 @@ sub argument_style {
 #     type arg
 #   PREINIT:
 #     aux vars
+#   [ALIAS:
+#     ID = INTEGER...]
 #   [PP]CODE:
 #     RETVAL = new Foo( THIS->method( arg1, *arg2 ) );
 #   POSTCALL:
@@ -208,14 +202,17 @@ sub print {
   my $args               = $this->arguments;
   my $ret_type           = $this->ret_type;
   my $ret_typemap        = $this->{TYPEMAPS}{RET_TYPE};
-  my $need_call_function = 0;
+  my $aliases            = $this->{ALIAS} || {};
+
+  my $has_aliases = scalar(keys %$aliases);
+
+  $out .= '#if ' . $this->emit_condition . "\n" if $this->emit_condition;
 
   my( $init, $arg_list, $call_arg_list, $code, $output, $cleanup,
-      $postcall, $precall ) =
-    ( '', '', '', '', '', '', '', '' );
-
-  my $use_ansi_style = $this->argument_style() eq 'ansi';
+      $postcall, $precall, $alias ) = ( ('') x 9 );
 
+  # compute the precall code, XS argument list and C++ argument list using
+  # the typemap information
   if( $args && @$args ) {
     my $has_self = $this->is_method ? 1 : 0;
     my( @arg_list, @call_arg_list );
@@ -225,14 +222,8 @@ sub print {
       my $pc  = $t->precall_code( sprintf( 'ST(%d)', $i + $has_self ),
                                   $arg->name );
 
-      $need_call_function ||=    defined $t->call_parameter_code( '' )
-                              || defined $pc;
-      my $type = $use_ansi_style ? $t->cpp_type . ' ' : '';
-      push @arg_list, $type . $arg->name .
+      push @arg_list, $t->cpp_type . ' ' . $arg->name .
                       ( $arg->has_default ? ' = ' . $arg->default : '' );
-      if (!$use_ansi_style) {
-        $init .= '    ' . $t->cpp_type . ' ' . $arg->name . "\n";
-      }
 
       my $call_code = $t->call_parameter_code( $arg->name );
       push @call_arg_list, defined( $call_code ) ? $call_code : $arg->name;
@@ -243,15 +234,14 @@ sub print {
     $call_arg_list = ' ' . join( ', ', @call_arg_list ) . ' ';
   }
 
-  # same for return value
-  $need_call_function ||= $ret_typemap &&
-    ( defined $ret_typemap->call_function_code( '', '' ) ||
-      defined $ret_typemap->output_code( '', '' ) ||
-      defined $ret_typemap->cleanup_code( '', '' ) );
-  # is C++ name != Perl name?
-  $need_call_function ||= $this->cpp_name ne $this->perl_name;
-  # package-static function
-  $need_call_function ||= $this->package_static;
+  # If there's %alias{foo = 123} definitions, generate ALIAS section
+  if ($has_aliases) {
+    # order by ordinal for consistent hash-order-independent output
+    my @alias_list = map "    $_ = $aliases->{$_}\n",
+                     sort {$aliases->{$a} <=> $aliases->{$b}}
+                     keys %$aliases;
+    $alias = "  ALIAS:\n" . join("", @alias_list);
+  }
 
   my $retstr = $ret_typemap ? $ret_typemap->cpp_type : 'void';
 
@@ -264,47 +254,51 @@ sub print {
 
   my $has_ret = $ret_typemap && !$ret_typemap->type->is_void;
 
-  # Hardcoded to one because we force the exception handling for now
-  # All the hard work above about determining whether $need_call_function
-  # needs to be enabled is left in as exception handling may be subject
-  # to configuration later. --Steffen
-  $need_call_function = 1;
-
   my $ppcode = $has_ret && $ret_typemap->output_list( '' ) ? 1 : 0;
   my $code_type = $ppcode ? "PPCODE" : "CODE";
-  if( $need_call_function ) {
-    my $ccode = $this->_call_code( $call_arg_list );
-    if ($this->isa('ExtUtils::XSpp::Node::Destructor')) {
-      $ccode = 'delete THIS';
-      $has_ret = 0;
-    } elsif( $has_ret && defined $ret_typemap->call_function_code( '', '' ) ) {
-      $ccode = $ret_typemap->call_function_code( $ccode, 'RETVAL' );
-    } elsif( $has_ret ) {
+  my $ccode = $this->_call_code( $call_arg_list );
+  if ($this->isa('ExtUtils::XSpp::Node::Destructor')) {
+    $ccode = 'delete THIS';
+    $has_ret = 0;
+  } elsif( $has_ret && defined $ret_typemap->call_function_code( '', '' ) ) {
+    $ccode = $ret_typemap->call_function_code( $ccode, 'RETVAL' );
+  } elsif( $has_ret ) {
+    if ($has_aliases) {
+      $ccode = $this->_generate_alias_conditionals($call_arg_list, 1); # 1 == use RETVAL
+    } else {
       $ccode = "RETVAL = $ccode";
     }
+  } elsif( $has_aliases ) { # aliases but no RETVAL
+    $ccode = $this->_generate_alias_conditionals($call_arg_list, 0); # 0 == no RETVAL
+  }
 
-    $code .= "  $code_type:\n";
-    $code .= "    try {\n";
-    $code .= '      ' . $precall if $precall;
-    $code .= '      ' . $ccode . ";\n";
-    if( $has_ret && defined $ret_typemap->output_code( '', '' ) ) {
-      $code .= '      ' . $ret_typemap->output_code( 'ST(0)', 'RETVAL' ) . ";\n";
-    }
-    if( $has_ret && defined $ret_typemap->output_list( '' ) ) {
-      $code .= '      ' . $ret_typemap->output_list( 'RETVAL' ) . ";\n";
-    }
-    $code .= "    }\n";
-    my @catchers = @{$this->{EXCEPTIONS}};
-    foreach my $exception_handler (@catchers) {
-      $code .= $exception_handler->handler_code;
-    }
+  $code .= "  $code_type:\n";
+  $code .= "    try {\n";
+  if ($precall) {
+    $code .= '      ' . $precall;
+  }
+  $code .= '      ' . $ccode . ";\n";
+  if( $has_ret && defined $ret_typemap->output_code( '', '' ) ) {
+    my $retcode = $ret_typemap->output_code( 'ST(0)', 'RETVAL' );
+    $code .= '      ' . $retcode . ";\n";
+  }
+  if( $has_ret && defined $ret_typemap->output_list( '' ) ) {
+    my $retcode = $ret_typemap->output_list( 'RETVAL' );
+    $code .= '      ' . $retcode . ";\n";
+  }
+  $code .= "    }\n";
+  my @catchers = @{$this->{EXCEPTIONS}};
+  foreach my $exception_handler (@catchers) {
+    my $handler_code = $exception_handler->handler_code;
+    $code .= $handler_code;
+  }
 
-    $output = "  OUTPUT: RETVAL\n" if $has_ret;
+  $output = "  OUTPUT: RETVAL\n" if $has_ret;
 
-    if( $has_ret && defined $ret_typemap->cleanup_code( '', '' ) ) {
-      $cleanup .= "  CLEANUP:\n";
-      $cleanup .= '    ' . $ret_typemap->cleanup_code( 'ST(0)', 'RETVAL' ) . ";\n";
-    }
+  if( $has_ret && defined $ret_typemap->cleanup_code( '', '' ) ) {
+    $cleanup .= "  CLEANUP:\n";
+    my $cleanupcode = $ret_typemap->cleanup_code( 'ST(0)', 'RETVAL' );
+    $cleanup .= '    ' . $cleanupcode . ";\n";
   }
 
   if( $this->code ) {
@@ -336,14 +330,30 @@ $cur_module PACKAGE=$pcname
 EOT
   }
 
-  $out .= "$retstr\n";
-  $out .= "$fname($arg_list)\n";
-  $out .= $init;
-  $out .= $code;
-  $out .= $postcall;
-  $out .= $output;
-  $out .= $cleanup;
-  $out .= "\n";
+  my $head = "$retstr\n"
+             . "$fname($arg_list)\n";
+  my $body = $alias . $init . $code . $postcall . $output . $cleanup . "\n";
+  $this->_munge_code(\$body) if $this->has_argument_with_length;
+
+  $out .= $head . $body;
+  $out .= '#endif // ' . $this->emit_condition . "\n" if $this->emit_condition;
+
+  return $out;
+}
+
+# This replaces the use of "length(varname)" with
+# the proper name of the XS variable that is auto-generated in
+# case of the XS length() feature. The Argument's take care of
+# this and do nothing if they're not of the "length" type.
+# Any additional checking "$this->_munge_code(\$code) if $using_length"
+# is just an optimization!
+sub _munge_code {
+  my $this = shift;
+  my $code = shift;
+  
+  foreach my $arg (@{$this->{ARGUMENTS}}) {
+    $$code = $arg->fix_name_in_code($$code);
+  }
 }
 
 =head2 print_declaration
@@ -378,9 +388,25 @@ but overridden in the L<ExtUtils::XSpp::Node::Method> sub-class.
 
 sub is_method { 0 }
 
+=head2 has_argument_with_length
+
+Returns true if the function has any argument that uses the XS length
+feature.
+
+=cut
+
+sub has_argument_with_length {
+  my $this = shift;
+  foreach my $arg (@{$this->{ARGUMENTS}}) {
+    return 1 if $arg->uses_length;
+  }
+  return();
+}
+
+
 =begin documentation
 
-ExtUtils::XSpp::Node::_call_code( argument_string )
+ExtUtils::XSpp::Function::_call_code( argument_string )
 
 Return something like "foo( $argument_string )".
 
@@ -390,6 +416,55 @@ Return something like "foo( $argument_string )".
 
 sub _call_code { return $_[0]->cpp_name . '(' . $_[1] . ')'; }
 
+=begin documentation
+
+ExtUtils::XSpp::Function::_call_code_aliased( function_alias_name, argument_string )
+
+Return something like "$function_alias_name( $argument_string )".
+
+=end documentation
+
+=cut
+
+sub _call_code_aliased { return $_[1] . '(' . $_[2] . ')'; }
+
+=begin documentation
+
+ExtUtils::XSpp::Function::_generate_alias_conditionals( argument_string, use_retval_bool )
+
+Generates if()else if()else block for XS function name aliasing (cf. the XS manual and the ix
+variable). If use_retval_bool is true, each included function call will contain an
+assignment to RETVAL.
+
+Returns the generated code.
+
+=end documentation
+
+=cut
+sub _generate_alias_conditionals {
+  my ($this, $call_arg_list, $use_retval) = @_;
+  my $aliases = $this->{ALIAS};
+
+  my $retval_code = $use_retval ? "RETVAL = " : "";
+  my $buf = "if (ix == 0) {\n  $retval_code"
+            . $this->_call_code($call_arg_list)
+            . ";\n}\n";
+  # order by ordinal for consistent hash-order-independent output
+  foreach my $alias (sort {$aliases->{$a} <=> $aliases->{$b}} keys %$aliases)
+  {
+    $buf .= "else if (ix == $aliases->{$alias}) {\n  "
+            . $retval_code . $this->_call_code_aliased($alias, $call_arg_list)
+            . ";\n}\n";
+  }
+  $buf .= "else\n  croak(\"Panic: Invalid invocation of function alias number %i!\", (int)ix))";
+
+  # indent
+  $buf =~ s/^/      /gm;
+  $buf =~ s/^\s+//; # first line will get special treatment...
+
+  return $buf;
+}
+
 
 =head1 ACCESSORS
 
@@ -436,6 +511,12 @@ Returns the C<%postcall> decorator if any.
 Returns the set of exception types that were associated
 with the function via C<%catch>. (array reference)
 
+=head2 aliases
+
+Returns a hashref of C<name =E<gt> position>
+function name aliases (see %alias and L<perlxs> ALIAS keyword).
+Does not include the main function name.
+
 =cut
 
 sub cpp_name { $_[0]->{CPP_NAME} }
@@ -449,12 +530,14 @@ sub set_code { $_[0]->{CODE} = $_[1] }
 sub cleanup { $_[0]->{CLEANUP} }
 sub postcall { $_[0]->{POSTCALL} }
 sub catch { $_[0]->{CATCH} ? $_[0]->{CATCH} : [] }
+sub aliases { $_[0]->{ALIAS} ? $_[0]->{ALIAS} : {} }
+sub tags { $_[0]->{TAGS} }
 
 =head2 set_static
 
 Sets the C<static>-ness attribute of the function.
-Can be either undef (i.e. not static), C<package_static>,
-or C<class_static>.
+Can be either undef (i.e. not static), C<"package_static">,
+or C<"class_static">.
 
 =head2 package_static
 
@@ -476,4 +559,50 @@ sub set_static { $_[0]->{STATIC} = $_[1] }
 sub package_static { ( $_[0]->{STATIC} || '' ) eq 'package_static' }
 sub class_static { ( $_[0]->{STATIC} || '' ) eq 'class_static' }
 
+=head2 ret_typemap
+
+Returns the typemap for the return value of the function.
+
+=head2 set_ret_typemap( typemap )
+
+Sets the typemap for the return value of the function.
+
+=head2 arg_typemap( index )
+
+Returns the typemap for one function arguments.
+
+=head2 set_arg_typemap( index, typemap )
+
+Sets the typemap for one function argument.
+
+=cut
+
+sub ret_typemap {
+  my ($this) = @_;
+
+  die "Typemap not available yet" unless $this->{TYPEMAPS}{RET_TYPE};
+  return $this->{TYPEMAPS}{RET_TYPE};
+}
+
+sub set_ret_typemap {
+  my ($this, $typemap) = @_;
+
+  $this->{TYPEMAPS}{RET_TYPE} = $typemap;
+}
+
+sub arg_typemap {
+  my ($this, $index) = @_;
+
+  die "Invalid index" unless $index < @{$this->{ARGUMENTS}};
+  die "Typemap not available yet" unless $this->{TYPEMAPS}{ARGUMENTS};
+  return $this->{TYPEMAPS}{ARGUMENTS}[$index];
+}
+
+sub set_arg_typemap {
+  my ($this, $index, $typemap) = @_;
+
+  die "Invalid index" unless $index < @{$this->{ARGUMENTS}};
+  $this->{TYPEMAPS}{ARGUMENTS}[$index] = $typemap;
+}
+
 1;
